classdef HelperBoundingBoxDetectorBlock < HelperBoundingBoxDetector
% HelperBoundingBoxDetectorBlock Bounding Box Detector for Lidar Tracking Example

% This is a helper class and may be removed or modified in a future
% release.

% Copyright 2019 The MathWorks, Inc.
    
    % Public nontunable properties for the block
    properties (Nontunable)
        % Name of the detection bus
        DetectionBusName = 'detectionBus'
        % Name of the segmentation information bus
        SegmentationBusName = 'segmentationBus'
        %NumLidarPoints Number of points in one Lidar scan
        NumLidarPoints = 55000;
    end
    
    % Store sample detection and maximum number defined by the bus name.
    properties (Nontunable, Access = protected)
        pSampleDetection
        pMaxNumDetections
    end
    
    methods(Access = protected)        
        function setupImpl(obj)
           % Store the sample detection by using the bus name
           obj.pSampleDetection =  struct('SensorIndex',1,...
               'Time',0,...
               'Measurement',zeros(7,1),...
               'MeasurementNoise',eye(7),...
               'ObjectClassID',0,...
               'MeasurementParameters',struct('OriginPosition',zeros(3,1)),...
               'ObjectAttributes',struct('TargetIndex',1));
           obj.pMaxNumDetections = 100;
        end
        
        function [detectionBus, time, segmentationBus] = stepImpl(obj, lidarScan)
            ptCloud = pointCloud(lidarScan);
            time = getCurrentTime(obj);
            [detections, obstacleIndices, groundIndices, croppedIndices] = stepImpl@HelperBoundingBoxDetector(obj,ptCloud, time);
            detectionBus = convertObjectDetectionsToBus(obj, detections);
            segmentationBus = convertIndicesToBus(obj, obstacleIndices, groundIndices, croppedIndices);
        end

        function icon = getIconImpl(~)
            % Define icon for System block
            icon = ["Helper","Bounding Box Detector"];
        end

        function inputName = getInputNamesImpl(~)
            % Return input port names for System block
            inputName = sprintf("Point Cloud\nLocations");
        end
        
        function outputName = getOutputNamesImpl(~)
           outputName = ["Detections";"Current Time";"Segmentation Information"]; 
        end
        
        function [out1, out2, out3] = getOutputSizeImpl(~)
            % Return size for each output port
            out1 = [1 1];
            out2 = [1 1];
            out3 = [1 1];
        end

        function [out1, out2, out3] = getOutputDataTypeImpl(obj)
            % Return data type for each output port
            out3 = obj.SegmentationBusName;
            out1 = obj.DetectionBusName;
            out2 = 'double';
        end

        function [out, out2, out3] = isOutputComplexImpl(~)
            % Return true for each output port with complex data
            out = false;
            out2 = false;
            out3 = false;
        end

        function [out, out2, out3] = isOutputFixedSizeImpl(~)
            % Return true for each output port with fixed size
            out = true;
            out2 = true;
            out3 = true;
        end
        
        function detectionBus = convertObjectDetectionsToBus(obj, detections)
            numDets = numel(detections);
            sampleStruct = obj.pSampleDetection;
            detStruct = repmat(sampleStruct,[obj.pMaxNumDetections 1]);
            for i = 1:numDets
                detStruct(i).SensorIndex = 1;
                detStruct(i).Time = detections{i}.Time;
                detStruct(i).Measurement = detections{i}.Measurement;
                detStruct(i).MeasurementNoise = detections{i}.MeasurementNoise;
            end
            detectionBus = struct('NumDetections',numDets,'Detections',detStruct);
        end
        
        function segmentationBus = convertIndicesToBus(obj, obstacleIndices, groundIndices, croppedIndices)
            logicalIndex = false(obj.NumLidarPoints,1);
            obstacleLogical = logicalIndex;
            obstacleLogical(obstacleIndices) = true;
            groundLogical = logicalIndex;
            groundLogical(groundIndices) = true;
            croppedLogical = logicalIndex;
            croppedLogical(croppedIndices) = true;
            segmentationBus = struct('ObstacleIndices', obstacleLogical,...
                'GroundIndices',groundLogical,...
                'CroppedIndices',croppedLogical);
        end
    end
    methods(Access = protected, Static)
        function header = getHeaderImpl
            % Define header panel for System block dialog
            header = matlab.system.display.Header('Title',mfilename,...
                'Text',HelperBoundingBoxDetectorBlock.descriptionText());
        end

        function simMode = getSimulateUsingImpl
            % Return only allowed simulation mode in System block
            % dialog
            simMode = ["Interpreted execution";"Code generation"];
        end
        
        function flag = showSimulateUsingImpl
            % Return false if simulation mode hidden in System block
            % dialog
            flag = true;
        end
        function text = descriptionText
            text = sprintf(['A helper block to segment the point cloud into bounding box detections. ',...
                    'In one step, the block process the point cloud in the following order: \n',...
                    '1. It removes the cloud outside defined limits.\n',...
                    '2. It segments the ground from the survived point cloud.\n',...
                    '3. It forms clusters from obstacle cloud and estimates a bounding box for each cluster.']);
        end
    end
end