classdef HelperLidarDataReader < matlab.System
    % Read Lidar data in Simulink.
    
    % This is a helper class and may be removed or modified in a future
    % release.
    
    % Copyright 2019 The MathWorks, Inc.
    properties (Nontunable)
        % TimeStep Time step between two lidar scans
        TimeStep = 0.1;
    end
    
    properties (Nontunable)
        % NumLidarPoints Number of lidar points in one scan
        NumLidarPoints = 55000
    end
    
    properties (Access = protected)
        Counter = 0;
    end
    
    properties (Access = protected)
        pLidarData
    end
    
    properties (Nontunable)
        % LidarFile MAT file for Lidar data
        LidarFile = {'lidarData_1','lidarData_2','lidarData_3','lidarData_4','lidarData_5'};
        
        % LidarDataVariableName Name of the lidar data in MAT file
        LidarDataVariableName = 'currentLidarData';
    end
    
    methods
        function obj = HelperLidarDataReader(varargin)
            setProperties(obj,nargin,varargin{:});
        end
    end
    
    methods (Access = protected)
        function setupImpl(obj)
            lidarData = cell(0,1);
            currentIndex = 0;
            for i = 1:numel(obj.LidarFile)
                s = load([obj.LidarFile{i},'.mat'],obj.LidarDataVariableName);
                thisLidar = s.(obj.LidarDataVariableName);
                numScans = numel(thisLidar);
                lidarData(currentIndex + (1:numScans)) = thisLidar;
                currentIndex = currentIndex + numScans;
            end
            obj.pLidarData = lidarData;
        end
        
        function lidarData = stepImpl(obj)
            time = getCurrentTime(obj);
            sts = getSampleTime(obj);
            dT = sts.SampleTime;
            counter = max(1,floor(time/dT));
            lidarData = obj.pLidarData{counter}.Location;
        end
        
        function out = getOutputNamesImpl(~)
           out = sprintf("Point Cloud\nLocations"); 
        end
        
        function out1 = getOutputSizeImpl(obj)
            out1 = [obj.NumLidarPoints 3];
        end
                    
        function d = getOutputDataTypeImpl(~)
           d = "single";
        end
        
        function varargout = isOutputComplexImpl(~)
            varargout = cell(nargout,1);
            for i = 1:nargout
                varargout{i} = false;
            end
        end
        
        function varargout = isOutputFixedSizeImpl(~)
            varargout = cell(nargout,1);
            for i = 1:nargout
                varargout{i} = true;
            end
        end

        function sts = getSampleTimeImpl(obj)
            % Define sample time type and parameters
            sts = obj.createSampleTime("Type", "Discrete", "SampleTime", obj.TimeStep);

            % Example: specify discrete sample time
            % sts = obj.createSampleTime("Type", "Discrete", ...
            %     "SampleTime", 1);
        end
        
        function icon = getIconImpl(~)
            icon = ["Helper","Lidar Data Reader"];
        end
        
    end

    methods(Access = protected, Static)
        function header = getHeaderImpl
            header = matlab.system.display.Header('Title','HelperLidarDataReader',...
                'Text',HelperLidarDataReader.descriptionText);
        end
        function simMode = getSimulateUsingImpl
            % Return only allowed simulation mode in System block dialog
            simMode = "Interpreted execution";
        end

        function flag = showSimulateUsingImpl
            % Return false if simulation mode hidden in System block dialog
            flag = false;
        end
        function text = descriptionText
           text = sprintf(['A helper block to read lidar data from a MAT file. ',...
               'The MAT file stores lidar data as a cell array of pointCloud objects. ',...
               'For lidar data split into multiple files, a cell array of MAT files can be provided. ',...
               'Each MAT file must contain lidar data with the same variable name. ',...
               'At each step, the block gets the current time from Simulink and uses the time-step between two lidar scans to ',...
               'compute the value of current scan.']);
        end
    end
end