classdef HelperLidarExampleDisplayBlock < HelperLidarExampleDisplay
   % HelperLidarExampleDisplayBlock A helper class for displaying 
   % lidar scan, detections and tracks in Simulink.
 
   properties (Nontunable)
      % LidarDataReaderBlockName Name of the lidar data reader block
      LidarDataReaderBlockName = 'Lidar Data Reader';
      
      % ImageDataReaderBlockName Name of the image data reader block
      ImageDataReaderBlockName = 'Image Data Reader';
      
      % DetectorBlockName Name of the bounding box detector block
      DetectorBlockName = 'Bounding Box Detector';
      
      % TrackerBlockName Name of the tracker block
      TrackerBlockName = 'Tracker';
   end
   
   % RunTimeObjects for each block to access data during simulation
   properties (Access = protected)
       pLidarDataReaderBlock
       pImageDataReaderBlock
       pTrackerBlock
       pDetectorBlock
   end
   
   methods
       function obj = HelperLidarExampleDisplayBlock(varargin)
           setProperties(obj,nargin,varargin{:});
%            obj.MovieName = 'a';
%            obj.RecordGIF = true;
       end
   end
   
   methods (Access = protected)
       function setupImpl(obj)
           % Setup the RunTimeObject for each block
           if isempty(obj.pLidarDataReaderBlock)
              obj.pLidarDataReaderBlock = get_param([gcs,'/',obj.LidarDataReaderBlockName],'RunTimeObject');
           end
           
           if isempty(obj.pImageDataReaderBlock)
              obj.pImageDataReaderBlock = get_param([gcs,'/',obj.ImageDataReaderBlockName],'RunTimeObject');
           end
           
           if isempty(obj.pTrackerBlock)
               obj.pTrackerBlock = get_param([gcs,'/',obj.TrackerBlockName],'RunTimeObject');
           end
           
           if isempty(obj.pDetectorBlock)
               obj.pDetectorBlock = get_param([gcs,'/',obj.DetectorBlockName],'RunTimeObject');
           end
           
           % Get current values of block outputs
           [referenceImage,lidarScan,segmentationInfo,detections,tracks] = getCurrentValues(obj);
 
           % Setup the display using base class
           detsIn = detections.Detections(1:detections.NumDetections);
           tracksIn = tracks.Tracks(1:tracks.NumTracks);
           obstacleIndices = segmentationInfo.ObstacleIndices;
           groundIndices = segmentationInfo.GroundIndices;
           croppedIndices = segmentationInfo.CroppedIndices;
           setupImpl@HelperLidarExampleDisplay(obj,detsIn,tracksIn,lidarScan,obstacleIndices,groundIndices,croppedIndices,referenceImage);
       end
       
       function stepImpl(obj)
           % Obtain the detections, tracks and segmentation information
           % from the blocks' RunTimeObjects
           [referenceImage,lidarScan,segmentationInfo,detections,tracks] = getCurrentValues(obj);
           
           % Parse and plot using base class
           detsIn = detections.Detections(1:detections.NumDetections);
           tracksIn = tracks.Tracks(1:tracks.NumTracks);
           obstacleIndices = segmentationInfo.ObstacleIndices;
           groundIndices = segmentationInfo.GroundIndices;
           croppedIndices = segmentationInfo.CroppedIndices;
           stepImpl@HelperLidarExampleDisplay(obj,detsIn,tracksIn,lidarScan,obstacleIndices,groundIndices,croppedIndices,referenceImage);
       end
       
       function inputNames = getInputNamesImpl(~)
           % Input names
          inputNames = ["Reference Image", sprintf("Point Cloud\nLocations"), sprintf("Segmentation\nInformation"), "Detections", "Tracks"];
       end
       
       function icon = getIconImpl(~)
          icon = ["Helper","Lidar Example Display"];
       end
       
       function flag = isInactivePropertyImpl(~,prop)
          flag = any(strcmpi(prop,{'Figure','PointCloudProcessingDisplay','TrackingDisplay','CloseUpDisplay','RecordGIF','MovieName'}));
       end
       
       function num = getNumInputsImpl(~)
          % No inputs
          num = 0;
       end
       
       function [referenceImage,lidarScan,segmentationInfo,detections,tracks] = getCurrentValues(obj)
           % Get current values of inputs using RunTimeObjects of the
           % blocks.
           referenceImage = obj.pImageDataReaderBlock.OutputPort(1).Data;
           lidarScan = obj.pLidarDataReaderBlock.OutputPort(1).Data;
           segmentationInfo = obj.pDetectorBlock.OutputPort(3).Data;
           detections = obj.pDetectorBlock.OutputPort(1).Data;
           tracks = obj.pTrackerBlock.OutputPort(1).Data;
       end
       
       function releaseImpl(obj)
%            writeAnimatedGIF(obj,1,40,'simulinkTracking','a');
       end
   end
   
   methods (Static, Access = protected)
       function header = getHeaderImpl
          header = matlab.system.display.Header('Title',mfilename,...
              'Text',HelperLidarExampleDisplayBlock.descriptionText);
       end
       function text = descriptionText
          text =  sprintf(['A helper class for displaying lidar scan, detections and tracks in Simulink. ',...
              'The block uses RunTimeObject parameters of the blocks to get their output values during simulation.',...
              'The display is divided into three main categories:\n',...
              '1. Lidar Preprocessing and Tracking: Lidar segmentation and tracking results.\n',...
              '2. Ego Vehicle Display: 2-D bird''s-eye view.\n',...
              '3. Tracking Details: Zoomed plot with estimated error covariances.\n']);
       end
       end
       
end